"use strict";
// Copyright (c) Frédéric Maquin <fred@ephread.com>
// Licensed under the MIT License.
// See LICENSE in the project root for license information.
Object.defineProperty(exports, "__esModule", { value: true });
const Progress = require("cli-progress");
const Extract = require("extract-zip");
const Fs = require("fs");
const Path = require("path");
const Request = require("request");
const INK_VERSION = '0.8.1';
/** Global progress used during archive download. */
let progress;
/** Messages displayed to the user. */
const messages = Object.freeze({
    unsupportedPlatform: "You are running a platform for which there are no prebuilt binaries. " +
        "The server comes bundled with an inklecate requiring the Mono runtime to run. " +
        "For more information, please read " +
        "https://github.com/ephread/ink-language-server/blob/master/README.md#inklecate",
    installError: "Inklecate could not be installed automatically. " +
        "For alternative options please read: " +
        "https://github.com/ephread/ink-language-server/blob/master/README.md#inklecate"
});
/**
 * Returns the name of the archive to download, depending on the architecture.
 */
function optionalMacOsBundleName() {
    const arch = process.arch;
    if (arch === "ia32" || arch === "x86") {
        return `inklecate_${INK_VERSION}-macos_i386.zip`;
    }
    if (arch === "x64") {
        return `inklecate_${INK_VERSION}-macos_x64.zip`;
    }
}
/**
 * Returns `true` if the script is running on MacOS.
 */
function isRunningOnMac() {
    return process.platform === "darwin";
}
/**
 * Returns `true` if the script is running on Windows.
 */
function isRunningOnWindows() {
    return process.platform === "win32";
}
/**
 * Returns `true` if the given HTTP response is Ok.
 */
function isResponseOk(response) {
    return response.statusCode >= 200 && response.statusCode < 300;
}
/**
 *
 *
 * @param error
 * @param callback
 */
function handleError(error, callback) {
    let message = "Cannot download inklecate. Reason: ";
    if (typeof error === "string") {
        message += error;
    }
    else {
        message += error.message;
    }
    callback(message);
}
/**
 * Download the prebuilt binary from the given `url`, and store it at the given `destinationPath`.
 *
 * @param url URL of the prebuilt binary.
 * @param destinationPath path at which store the binary.
 * @param callback callback called upon completion of the download or when an error is encountered.
 */
function downloadDependency(url, destinationPath, callback) {
    const request = Request(url).on("response", response => {
        if (!isResponseOk(response)) {
            handleError(`HTTP error ${response.statusCode} ${response.statusMessage}`, callback);
        }
        else {
            request
                .pipe(Fs.createWriteStream(destinationPath).on("error", callback))
                .on("finish", callback);
            const length = parseInt(response.headers["content-length"], 10);
            progress = new Progress.Bar({}, Progress.Presets.legacy);
            progress.start(length, 0);
            response
                .on("data", chunk => {
                progress.increment(chunk.length);
            })
                .on("end", () => {
                progress.stop();
            });
        }
    }).on("error", error => {
        handleError(`${error}`, callback);
    });
}
/**
 *
 */
function checkPlatformAndDownloadBinaryDependency(connection, callback) {
    if (isRunningOnMac()) {
        // Running on a mac.
        const bundleName = optionalMacOsBundleName();
        const urlpart = `https://dl.bintray.com/ephread/ink-language-server/`;
        if (bundleName) {
            const url = `${urlpart}${bundleName}`;
            const vendorDir = Path.join(__dirname, "../vendor/");
            const filePath = Path.join(vendorDir, bundleName);
            Fs.stat(Path.join(vendorDir, 'inklecate'), (statError, stat) => {
                if (stat) {
                    connection.log(`Running on macOS, inklecate has already been downloaded.`);
                    callback(true);
                }
                else {
                    connection.log(`Running on macOS, fetching inkeclate from… ${url}`);
                    downloadDependency(url, filePath, error => {
                        if (error) {
                            if (progress) {
                                progress.stop();
                            }
                            connection.log(`${error}`);
                            connection.log(messages.installError);
                            callback(false);
                        }
                        else {
                            Extract(filePath, { dir: vendorDir }, extractError => {
                                if (extractError) {
                                    connection.log(extractError.message);
                                    callback(false);
                                }
                                else {
                                    Fs.unlink(filePath, unLinkError => {
                                        if (unLinkError) {
                                            connection.log(`${unLinkError}`);
                                        }
                                    });
                                    connection.log("Inklecate successfully installed!");
                                    callback(true);
                                }
                            });
                        }
                    });
                }
            });
        }
    }
    else if (!isRunningOnWindows()) {
        connection.log(messages.unsupportedPlatform);
        callback(true);
    }
    else {
        callback(true);
    }
}
exports.checkPlatformAndDownloadBinaryDependency = checkPlatformAndDownloadBinaryDependency;
//# sourceMappingURL=install.js.map