"use strict";
// Copyright (c) Frédéric Maquin <fred@ephread.com>
// Licensed under the MIT License.
// See LICENSE in the project root for license information.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const inklecate_1 = require("./inklecate");
const types_1 = require("./types");
const configuration_1 = require("./configuration");
const identifiers_1 = require("./identifiers");
const utils_1 = require("./utils");
const install_1 = require("./install");
const story_renderer_1 = require("./story-renderer");
/* Properties */
/******************************************************************************/
/** Capabilities defined by the client. Defaults to none. */
const capabilities = {
    configuration: false,
    workspaceFolder: false,
    diagnostic: false
};
/** Connection with the client. */
const connection = vscode_languageserver_1.createConnection(vscode_languageserver_1.ProposedFeatures.all);
const logger = new types_1.InkConnectionLogger(connection);
/** Documents managed by the clients. */
const documents = new vscode_languageserver_1.TextDocuments();
/**
 * Workspace directories managed by the server,
 * mapped from the workspace managed by the client.
 *
 * URIs of the client's `WorkspaceFolder` will be used as keys.
 */
const workspaceDirectories = new Map();
/**
 * Document settings, scoped per document, returned by the client.
 *
 * URIs of the client's `TextDocument` will be used as keys.
 */
const documentSettings = new Map();
let canCompile = false;
/* Helpers */
/******************************************************************************/
/**
 * Returns the InkWorkspace of the given document. If no workspace could be found,
 * returns `undefined`.
 *
 * Since Inklecate requires as a full compilation of the entire related project
 * to push diagnostics, the workspace of each document must be known.
 *
 * @param document a text document, belonging to the returned workspace.
 */
function getInkWorkspaceOfDocument(document) {
    const documentPath = vscode_uri_1.default.parse(document.uri).fsPath;
    return getInkWorkspaceOfFilePath(documentPath);
}
/**
 * Returns the InkWorkspace of the given uri. If no workspace could be
 * found, returns `undefined`.
 *
 * Since Inklecate requires as a full compilation of the entire related project
 * to push diagnostics, the workspace of each document must be known.
 *
 * @param document a text document, belonging to the returned workspace.
 */
function getInkWorkspaceOfFilePath(documentUri) {
    for (const workspaceKey of workspaceDirectories.keys()) {
        const workspace = workspaceDirectories.get(workspaceKey);
        if (!workspace) {
            continue;
        }
        const workspacePath = vscode_uri_1.default.parse(workspace.folder.uri).fsPath;
        if (utils_1.isFilePathChildOfDirPath(documentUri, workspacePath)) {
            return workspace;
        }
    }
    return undefined;
}
/**
 * Fetch the configuration settings of the given document, from the client.
 *
 * @param document the document of which fetch the configuration settings,
 *                 can either be a TextDocument or a string-based uri.
 */
function fetchDocumentConfigurationSettings(documentOrUri) {
    let documentUri;
    if (typeof documentOrUri === "string") {
        documentUri = documentOrUri;
    }
    else {
        documentUri = documentOrUri.uri;
    }
    if (!capabilities.configuration) {
        return Promise.resolve(Object.assign({}, configuration_1.getDefaultSettings()));
    }
    let result = documentSettings.get(documentUri);
    if (!result) {
        result = connection.workspace.getConfiguration({
            scopeUri: documentUri,
            section: "ink"
        });
        documentSettings.set(documentUri, result);
    }
    return result;
}
/**
 * Send the given errors to the client.
 *
 * @param errors the errors to push.
 */
function notifyClientAndPushDiagnostics(workspace, outputStoryPath, errors) {
    for (const textDocument of documents.all()) {
        const diagnostics = [];
        for (const error of errors) {
            if (vscode_uri_1.default.parse(textDocument.uri).fsPath === error.filePath) {
                let message = error.message;
                if (error.type === types_1.InkErrorType.Todo) {
                    message = `Todo: ${message}`;
                }
                const diagnostic = {
                    severity: utils_1.getDiagnosticSeverityFromInkErrorType(error.type),
                    range: {
                        start: { line: error.lineNumber - 1, character: 0 },
                        end: { line: error.lineNumber, character: 0 }
                    },
                    message,
                    source: "inklecate"
                };
                diagnostics.push(diagnostic);
            }
        }
        // If there a no errors to report, we'll send a custom notification to
        // the client, asking it to remember the path to the compiled story.
        if (diagnostics.length === 0) {
            const params = {
                workspaceUri: workspace.folder.uri,
                storyUri: `file://${outputStoryPath}`
            };
            connection.sendNotification(identifiers_1.CompilationNotification.didCompileStory, params);
        }
        connection.sendDiagnostics({ uri: textDocument.uri, diagnostics });
    }
}
/**
 * Initialize workspaces by fetching opened `WorkspaceFolder` from the client and
 * creating the temporary directories which will hold copies of the Ink Project.
 */
function initializeInkWorkspaces() {
    return __awaiter(this, void 0, void 0, function* () {
        const workspaceFolders = yield connection.workspace.getWorkspaceFolders();
        if (workspaceFolders) {
            for (const workspaceFolder of workspaceFolders) {
                let workspace = workspaceDirectories.get(workspaceFolder.uri);
                if (!workspace) {
                    workspace = {
                        folder: workspaceFolder
                    };
                    workspaceDirectories.set(workspaceFolder.uri, workspace);
                }
                inklecate_1.prepareTempDirectoryForCompilation(workspaceFolder, logger, tempDir => {
                    if (tempDir) {
                        workspace.temporaryCompilationDirectory = tempDir;
                    }
                    else {
                        reportServerError();
                    }
                });
            }
        }
    });
}
/**
 * Update the content of the given document, in the temporary copy of the workspace,
 * and compile the project.
 *
 * @param document the document to update.
 */
function updateDocumentAndCompileWorkspace(document) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!canCompile) {
            return;
        }
        const workspace = getInkWorkspaceOfDocument(document);
        if (!workspace) {
            return;
        }
        const settings = yield fetchDocumentConfigurationSettings(document);
        inklecate_1.updateFile(document, workspace, logger, error => {
            if (error) {
                connection.console.log(`Could not update '${document.uri}', ${error.message}`);
                reportServerError();
            }
            else {
                inklecate_1.compileProject(settings, workspace, logger, undefined, notifyClientAndPushDiagnostics);
            }
        });
    });
}
/**
 * Compile the story contained in `workspace`.
 * `documentPath` is required to retrieve the proper settings, which
 * are scoped to the resource.
 *
 * @param documentUri the document URI used to retrieved the settings.
 * @param workspace the ink workspace to compile.
 */
function executeCompileCommand(documentUri, workspace, play = false) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!canCompile) {
            return;
        }
        const settings = yield fetchDocumentConfigurationSettings(documentUri);
        const storyRenderer = play ? new story_renderer_1.StoryRenderer(connection) : undefined;
        inklecate_1.compileProject(settings, workspace, logger, storyRenderer, notifyClientAndPushDiagnostics);
    });
}
function reportServerError() {
    connection.window.showErrorMessage("An unrecoverable error occured with Ink Language Server. " +
        "Please see the logs for more information.");
}
/* Connection callbacks */
/******************************************************************************/
connection.onInitialize((params) => {
    connection.console.log("Server Initialized");
    const clientCapabilities = params.capabilities;
    if (clientCapabilities.workspace) {
        capabilities.configuration = !!clientCapabilities.workspace.configuration;
        capabilities.workspaceFolder = !!clientCapabilities.workspace.workspaceFolders;
        if (clientCapabilities.textDocument) {
            capabilities.diagnostic = !!clientCapabilities.textDocument.publishDiagnostics;
        }
    }
    return {
        capabilities: {
            textDocumentSync: documents.syncKind,
            executeCommandProvider: {
                commands: [
                    identifiers_1.Commands.compileStory,
                    identifiers_1.Commands.playStory,
                    identifiers_1.Commands.selectOption
                ]
            }
        }
    };
});
connection.onInitialized(() => {
    if (capabilities.configuration) {
        connection.client.register(vscode_languageserver_1.DidChangeConfigurationNotification.type, undefined);
    }
    if (capabilities.workspaceFolder) {
        connection.workspace.onDidChangeWorkspaceFolders(event => {
            initializeInkWorkspaces();
        });
    }
    initializeInkWorkspaces();
    install_1.checkPlatformAndDownloadBinaryDependency(logger, (success) => {
        configuration_1.flagDefaultSettingsAsDirty();
        canCompile = success;
    });
});
connection.onDidChangeConfiguration(change => {
    documentSettings.clear();
});
connection.onExecuteCommand((params) => {
    switch (params.command) {
        case identifiers_1.Commands.compileStory:
            runCompileStoryCommand(params);
            break;
        case identifiers_1.Commands.playStory:
            runPlayStoryCommand(params);
            break;
        case identifiers_1.Commands.selectOption:
            runSelectOptionCommand(params);
            break;
        default: break;
    }
});
/* Document callbacks */
/******************************************************************************/
documents.onDidChangeContent(change => {
    updateDocumentAndCompileWorkspace(change.document);
});
documents.onDidClose(event => {
    documentSettings.delete(event.document.uri);
});
/* Final setup */
/******************************************************************************/
documents.listen(connection);
connection.listen();
function runCompileStoryCommand(params) {
    connection.console.log('Received compilation command.');
    const pathAndWorkspace = getDocumentPathFromParams(params);
    if (!pathAndWorkspace) {
        return undefined;
    }
    executeCompileCommand(pathAndWorkspace.documentPath, pathAndWorkspace.workspace);
}
function runPlayStoryCommand(params) {
    connection.console.log('Received play story command.');
    const pathAndWorkspace = getDocumentPathFromParams(params);
    if (!pathAndWorkspace) {
        return undefined;
    }
    executeCompileCommand(pathAndWorkspace.documentPath, pathAndWorkspace.workspace, true);
}
function runSelectOptionCommand(params) {
    connection.console.log('Received select option command.');
    if (!params.arguments || !params.arguments[0]) {
        connection.window.showErrorMessage(`${identifiers_1.Commands.selectOption} error: the command was called with no arguments.`);
        return;
    }
    const index = parseInt(params.arguments[0]);
    inklecate_1.chooseOption(index);
}
function runKillInklecateCommand() {
    connection.console.log('Received kill inklecate command.');
    inklecate_1.killInklecate();
}
function getDocumentPathFromParams(params) {
    let fileURI;
    if (!params.arguments || params.arguments.length < 1) {
        fileURI = configuration_1.getDefaultSettings().mainStoryPath;
    }
    else {
        fileURI = params.arguments[0];
    }
    const documentPath = vscode_uri_1.default.parse(fileURI).fsPath;
    const workspace = getInkWorkspaceOfFilePath(documentPath);
    if (!workspace) {
        connection.console.log("Could not retrieve the workspace of the given file.");
        return undefined;
    }
    return { documentPath, workspace };
}
//# sourceMappingURL=server.js.map